function [W, loss] = GLassoUpdate(fLoss, fGradient, W, lambda, maxFunEvals)

if (nargin < 5) || isempty(maxFunEvals), maxFunEvals = 20; end;

TolW = 1e-5;

%
% LASSO phase: Fix free weights, only update regularized weights.
%

%
% Gradient LASSO - Addition step.
%

% Compute the gradient for the current parameters (row vectors)
G = fGradient(W);

% Find best search dimension k and sign gamma: argmin gamma*Gradient(W)_k
[minval, k] = max(abs(G));
gamma_k = -sign(G(k));

% Form search vector.
V = zeros(size(W));
V(k) = gamma_k * lambda;

% Line search for alpha=[0,1] argmin C(W[alpha, v]).
[alpha, loss] = fminbnd(@(a) fLoss((1-a).*W+a.*V), 0, 1, ...
  optimset('Display', 'notify', 'MaxFunEvals', maxFunEvals));
if (alpha > 0.999) && (fLoss(V) <= loss)
  alpha = 1;
end

% Update W.
W = (1-alpha).*W+alpha.*V;
disp(sprintf('After adding %d, |W|=%f', [k, sum(abs(W))]));

% Recompute gradient for the deletion step after the update.
G = fGradient(W);


%
% Gradient LASSO - Deletion step.
%

% Select non-zero parameters.
% TODO: Add tolerance to this!
I_nonzero = find(abs(W) > TolW);
W_nonzero = W(I_nonzero);
G_nonzero = G(I_nonzero);
S_nonzero = sign(W_nonzero);

projection = G_nonzero*S_nonzero';
isOnBoundary = (sum(abs(W_nonzero)) >= lambda - TolW);
if (projection < 0) && (isOnBoundary)
  % projection < 0 => moving in -G direction would increase |W|.
  % |W| = lambda => we are on the boundary. We need to move on the edge.
  % H_sigma = -Gradient_sigma + <Theta,Gradient>/|Sigma|*Theta
  H_nonzero = -G_nonzero + (projection ./ size(I_nonzero, 2)) .* S_nonzero;
else
  H_nonzero = -G_nonzero;
end

% Form search vector.
H = zeros(size(W));
H(I_nonzero) = H_nonzero;

% Compute maximum step size.
if (isOnBoundary)
  I_L = find((W_nonzero .* H_nonzero) < 0); % Weights that want to decrease.
  if ~isempty(I_L)
    [L, k] = min(-W_nonzero(I_L) ./ H_nonzero(I_L));
    disp(sprintf('Given L=%f ; deleting %d', [L, I_nonzero(I_L(k))]));
  else % isempty(I_L)
    % No feasible deletions.
    disp('GLasso: No feasible deletions!');
    return
  end
else
  absH = sum(abs(H_nonzero));
  if (absH >= 1e-6)
    L = (lambda - sum(abs(W_nonzero)))/absH;
    disp(sprintf('My L=%f', L));
  else % |H| almost zero.
    disp(sprintf('|H|=%f -- optimum reached inside simplex!', absH));
    return
  end
end


% Line search for delta=(0,L) argmin C(W+delta*H)
[alpha, loss] = fminbnd(@(d) fLoss(W+d.*L.*H), 0, 1, ...
  optimset('Display', 'notify', 'MaxFunEvals', maxFunEvals));
if (alpha > 0.999) && (fLoss(W+L.*H) <= loss)
  alpha = 1;
end

disp(sprintf('Deletion step size: %f', alpha * L));

% Update W=W+delta*H
W = W + alpha * L .* H;
disp(sprintf('After deletion, |W|=%f', sum(abs(W))));
  
return
